from unittest import TestCase

from servicelayer.cache import get_fakeredis
from servicelayer.rate_limit import RateLimit


class RateLimitTest(TestCase):
    def test_rate(self):
        conn = get_fakeredis()
        limit = RateLimit(conn, "banana", limit=10)
        assert limit.check()
        limit.update()
        assert limit.check()
        for num in range(13):
            assert num + 2 == limit.update()
        assert not limit.check()



import datetime
from unittest import TestCase
from unittest.mock import patch
import json
from random import randrange
import time_machine

import pika
from prometheus_client import REGISTRY
from prometheus_client.metrics import MetricWrapperBase
import pytest

from servicelayer import settings
from servicelayer.cache import get_fakeredis
from servicelayer.taskqueue import (
    Worker,
    Dataset,
    Task,
    get_rabbitmq_channel,
    dataset_from_collection_id,
    declare_rabbitmq_queue,
    flush_queues,
)
from servicelayer.util import unpack_datetime

from servicelayer.taskqueue import get_priority, get_task_count, queue_task


class CountingWorker(Worker):
    def dispatch_task(self, task: Task) -> Task:
        assert isinstance(task, Task), task
        if not hasattr(self, "test_done"):
            self.test_done = 0
        self.test_done += 1
        self.test_task = task
        return task


class FailingWorker(Worker):
    def dispatch_task(self, task: Task) -> Task:
        raise Exception("Woops")


class TaskQueueTest(TestCase):
    def test_task_queue(self):
        test_queue_name = "sls-queue-ingest"
        conn = get_fakeredis()
        conn.flushdb()
        collection_id = 2
        task_id = "test-task"
        priority = randrange(1, settings.RABBITMQ_MAX_PRIORITY + 1)
        body = {
            "collection_id": 2,
            "job_id": "test-job",
            "task_id": "test-task",
            "operation": "test-op",
            "context": {},
            "payload": {},
            "priority": priority,
        }
        channel = get_rabbitmq_channel()
        declare_rabbitmq_queue(channel, test_queue_name)
        channel.queue_purge(test_queue_name)
        channel.basic_publish(
            properties=pika.BasicProperties(priority=priority),
            exchange="",
            routing_key=test_queue_name,
            body=json.dumps(body),
        )
        dataset = Dataset(conn=conn, name=dataset_from_collection_id(collection_id))
        dataset.add_task(task_id, "test-op")
        channel.close()
        status = dataset.get_status()
        assert status["finished"] == 0, status
        assert status["pending"] == 1, status
        assert status["running"] == 0, status
        started = unpack_datetime(status["start_time"])
        last_updated = unpack_datetime(status["last_update"])
        assert started < last_updated
        assert abs(started - last_updated) < datetime.timedelta(seconds=1)

        worker = CountingWorker(queues=[test_queue_name], conn=conn, num_threads=1)
        worker.process(blocking=False)

        status = dataset.get_status()
        assert status["finished"] == 0, status
        assert status["pending"] == 0, status
        assert status["running"] == 1, status
        assert worker.test_done == 1
        task = Task(**body, delivery_tag=None)
        assert task.get_retry_count(conn) == 1

        with patch("servicelayer.settings.WORKER_RETRY", 0):
            channel = get_rabbitmq_channel()
            channel.queue_purge(test_queue_name)
            channel.basic_publish(
                properties=pika.BasicProperties(priority=priority),
                exchange="",
                routing_key=test_queue_name,
                body=json.dumps(body),
            )
            dataset = Dataset(conn=conn, name=dataset_from_collection_id(collection_id))
            dataset.add_task(task_id, "test-op")
            channel.close()
            with self.assertLogs(level="ERROR") as ctx:
                with patch.object(
                    pika.channel.Channel,
                    attribute="basic_nack",
                    return_value=None,
                ) as nack_fn:
                    worker.process(blocking=False)
                    nack_fn.assert_any_call(
                        delivery_tag=1, multiple=False, requeue=False
                    )
            assert "Max retries reached for task test-task. Aborting." in ctx.output[0]
            # Assert that the retry count is 0 and the key is gone
            assert task.get_retry_count(conn) == 0
            assert conn.get("tq:qdj:2:taskretry:test-task") is None

        worker.ack_message(worker.test_task, channel)
        status = dataset.get_status()
        assert status["finished"] == 0, status
        assert status["pending"] == 0, status
        assert status["running"] == 0, status
        assert status["start_time"] is None
        assert status["last_update"] is None

    @patch("servicelayer.taskqueue.Dataset.should_execute")
    def test_task_that_shouldnt_execute(self, mock_should_execute):
        test_queue_name = "sls-queue-ingest"
        conn = get_fakeredis()
        conn.flushdb()
        collection_id = 2
        task_id = "test-task"
        priority = randrange(1, settings.RABBITMQ_MAX_PRIORITY + 1)
        body = {
            "task_id": "test-task",
            "job_id": "test-job",
            "delivery_tag": 0,
            "operation": "test-op",
            "context": {},
            "payload": {},
            "priority": priority,
            "collection_id": 2,
        }

        channel = get_rabbitmq_channel()
        declare_rabbitmq_queue(channel, test_queue_name)
        channel.queue_purge(test_queue_name)
        channel.basic_publish(
            properties=pika.BasicProperties(priority=priority),
            exchange="",
            routing_key=test_queue_name,
            body=json.dumps(body),
        )

        def did_nack():
            return False

        channel.add_on_return_callback(did_nack)

        mock_should_execute.return_value = False

        dataset = Dataset(conn=conn, name=dataset_from_collection_id(collection_id))
        dataset.add_task(task_id, "test-op")
        status = dataset.get_active_dataset_status(conn=conn)
        stage = status["datasets"]["2"]["stages"][0]
        assert stage["pending"] == 1
        assert stage["running"] == 0

        worker = CountingWorker(queues=[test_queue_name], conn=conn, num_threads=1)
        assert not dataset.should_execute(task_id=task_id)
        with patch.object(
            worker,
            attribute="dispatch_task",
            return_value=None,
        ) as dispatch_fn:
            with patch.object(
                channel,
                attribute="basic_nack",
                return_value=None,
            ) as nack_fn:
                worker.process(blocking=False)
                nack_fn.assert_called_once()
                dispatch_fn.assert_not_called()

        status = dataset.get_active_dataset_status(conn=conn)
        stage = status["datasets"]["2"]["stages"][0]
        assert stage["pending"] == 1
        assert stage["running"] == 0
        assert dataset.is_task_tracked(Task(**body))


def test_dataset_get_status():
    conn = get_fakeredis()
    conn.flushdb()

    dataset = Dataset(conn=conn, name="123")
    status = dataset.get_status()

    assert status["pending"] == 0
    assert status["running"] == 0
    assert status["finished"] == 0
    assert status["start_time"] is None
    assert status["last_update"] is None

    task_one = Task(
        task_id="1",
        job_id="abc",
        delivery_tag="",
        operation="ingest",
        context={},
        payload={},
        priority=5,
        collection_id="1",
    )

    task_two = Task(
        task_id="2",
        job_id="abc",
        delivery_tag="",
        operation="ingest",
        context={},
        payload={},
        priority=5,
        collection_id="1",
    )

    task_three = Task(
        task_id="3",
        job_id="abc",
        delivery_tag="",
        operation="index",
        context={},
        payload={},
        priority=5,
        collection_id="1",
    )

    # Adding a task updates `start_time` and `last_update`
    with time_machine.travel("2024-01-01T00:00:00"):
        dataset.add_task(task_one.task_id, task_one.operation)

    status = dataset.get_status()
    assert status["pending"] == 1
    assert status["running"] == 0
    assert status["finished"] == 0
    assert status["start_time"].startswith("2024-01-01T00:00:00")
    assert status["last_update"].startswith("2024-01-01T00:00:00")

    # Once a worker starts processing a task, only `last_update` is updated
    with time_machine.travel("2024-01-02T00:00:00"):
        dataset.checkout_task(task_one.task_id, task_one.operation)

    status = dataset.get_status()
    assert status["pending"] == 0
    assert status["running"] == 1
    assert status["finished"] == 0
    assert status["start_time"].startswith("2024-01-01T00:00:00")
    assert status["last_update"].startswith("2024-01-02T00:00:00")

    # When another task is added, only `last_update` is updated
    with time_machine.travel("2024-01-03T00:00:00"):
        dataset.add_task(task_two.task_id, task_two.operation)

    status = dataset.get_status()
    assert status["pending"] == 1
    assert status["running"] == 1
    assert status["finished"] == 0
    assert status["start_time"].startswith("2024-01-01T00:00:00")
    assert status["last_update"].startswith("2024-01-03T00:00:00")

    # When the first task has been processed, `last_update` is updated
    with time_machine.travel("2024-01-04T00:00:00"):
        dataset.mark_done(task_one)

    status = dataset.get_status()
    assert status["pending"] == 1
    assert status["running"] == 0
    assert status["finished"] == 1
    assert status["start_time"].startswith("2024-01-01T00:00:00")
    assert status["last_update"].startswith("2024-01-04T00:00:00")

    # When the worker starts processing the second task, only `last_update` is updated
    with time_machine.travel("2024-01-05T00:00:00"):
        dataset.checkout_task(task_two.task_id, task_two.operation)

    status = dataset.get_status()
    assert status["pending"] == 0
    assert status["running"] == 1
    assert status["finished"] == 1
    assert status["start_time"].startswith("2024-01-01T00:00:00")
    assert status["last_update"].startswith("2024-01-05T00:00:00")

    # Once all tasks have been processed, status data is flushed
    with time_machine.travel("2024-01-06T00:00:00"):
        dataset.mark_done(task_two)

    status = dataset.get_status()
    assert status["pending"] == 0
    assert status["running"] == 0
    assert status["finished"] == 0
    assert status["start_time"] is None
    assert status["last_update"] is None

    # Adding a new task to an inactive dataset sets `start_time`
    with time_machine.travel("2024-01-07T00:00:00"):
        dataset.add_task(task_three.task_id, task_three.operation)

    status = dataset.get_status()
    assert status["pending"] == 1
    assert status["running"] == 0
    assert status["finished"] == 0
    assert status["start_time"].startswith("2024-01-07T00:00:00")
    assert status["last_update"].startswith("2024-01-07T00:00:00")

    # Cancelling a dataset flushes status data
    with time_machine.travel("2024-01-08T00:00:00"):
        dataset.checkout_task(task_three.task_id, task_three.operation)
        dataset.cancel()

    status = dataset.get_status()
    assert status["pending"] == 0
    assert status["running"] == 0
    assert status["finished"] == 0
    assert status["start_time"] is None
    assert status["last_update"] is None

    # Tasks that were already running when the dataset was cancelled
    # have no effect
    with time_machine.travel("2024-01-09T00:00:00"):
        dataset.mark_done(task_three)

    assert status["pending"] == 0
    assert status["running"] == 0
    assert status["finished"] == 0
    assert status["start_time"] is None
    assert status["last_update"] is None


def test_dataset_cancel():
    conn = get_fakeredis()
    conn.flushdb()

    dataset = Dataset(conn=conn, name="abc")
    assert conn.keys() == []

    # Enqueueing tasks stores status data in Redis
    dataset.add_task("1", "ingest")
    dataset.add_task("2", "index")
    dataset.checkout_task("1", "ingest")
    assert conn.keys() != []

    # Cancelling a dataset removes associated data from Redis
    dataset.cancel()
    assert conn.keys() == []


def test_dataset_mark_done():
    conn = get_fakeredis()
    conn.flushdb()

    dataset = Dataset(conn=conn, name="abc")
    assert conn.keys() == []

    task = Task(
        task_id="1",
        job_id="abc",
        delivery_tag="",
        operation="ingest",
        context={},
        payload={},
        priority=5,
        collection_id="abc",
    )

    # Enqueueing a task stores status data in Redis
    dataset.add_task(task.task_id, task.operation)
    dataset.checkout_task(task.task_id, task.operation)
    assert conn.keys() != []

    # Marking the last task as done cleans up status data in Redis
    dataset.mark_done(task)
    assert conn.keys() == []


@pytest.fixture
def prom_registry():
    # This relies on internal implementation details of the client to reset
    # previously collected metrics before every test execution. Unfortunately,
    # there is no clean way of achieving the same thing that doesn't add a lot
    # of complexity to the test and application code.
    collectors = REGISTRY._collector_to_names.keys()
    for collector in collectors:
        if isinstance(collector, MetricWrapperBase):
            collector._metrics.clear()
            collector._metric_init()

    yield REGISTRY


def test_prometheus_metrics_succeeded(prom_registry):
    conn = get_fakeredis()
    rmq_channel = get_rabbitmq_channel()
    worker = CountingWorker(conn=conn, queues=["ingest"], num_threads=1)
    declare_rabbitmq_queue(channel=rmq_channel, queue="ingest")

    queue_task(
        rmq_channel=rmq_channel,
        redis_conn=conn,
        collection_id=123,
        stage="ingest",
    )
    worker.process(blocking=False)

    started = prom_registry.get_sample_value(
        "servicelayer_tasks_started_total",
        {"stage": "ingest"},
    )
    assert started == 1

    succeeded = prom_registry.get_sample_value(
        "servicelayer_tasks_succeeded_total",
        {"stage": "ingest", "retries": "0"},
    )
    assert succeeded == 1

    # Under the hood, histogram metrics create multiple time series tracking
    # the number and sum of observations, as well as individual histogram buckets.
    duration_sum = prom_registry.get_sample_value(
        "servicelayer_task_duration_seconds_sum",
        {"stage": "ingest"},
    )
    duration_count = prom_registry.get_sample_value(
        "servicelayer_task_duration_seconds_count",
        {"stage": "ingest"},
    )
    assert duration_sum > 0
    assert duration_count == 1


def test_prometheus_metrics_failed(prom_registry):
    conn = get_fakeredis()
    rmq_channel = get_rabbitmq_channel()
    worker = FailingWorker(conn=conn, queues=["ingest"], num_threads=1)
    declare_rabbitmq_queue(channel=rmq_channel, queue="ingest")

    queue_task(
        rmq_channel=rmq_channel,
        redis_conn=conn,
        collection_id=123,
        stage="ingest",
    )
    worker.process(blocking=False)

    started = prom_registry.get_sample_value(
        "servicelayer_tasks_started_total",
        {"stage": "ingest"},
    )
    assert settings.WORKER_RETRY == 3
    assert started == 4  # Initial attempt + 3 retries

    first_attempt = REGISTRY.get_sample_value(
        "servicelayer_tasks_failed_total",
        {
            "stage": "ingest",
            "retries": "0",
            "failed_permanently": "False",
        },
    )
    assert first_attempt == 1

    last_attempt = REGISTRY.get_sample_value(
        "servicelayer_tasks_failed_total",
        {
            "stage": "ingest",
            "retries": "3",
            "failed_permanently": "True",
        },
    )
    assert last_attempt == 1


def test_get_priority_bucket():
    redis = get_fakeredis()
    rmq_channel = get_rabbitmq_channel()
    rmq_channel.queue_delete("index")
    declare_rabbitmq_queue(rmq_channel, "index")
    flush_queues(rmq_channel, redis, ["index"])
    collection_id = 1

    assert get_task_count(collection_id, redis) == 0
    assert get_priority(collection_id, redis) in (7, 8)

    queue_task(rmq_channel, redis, collection_id, "index")

    assert get_task_count(collection_id, redis) == 1
    assert get_priority(collection_id, redis) in (7, 8)

    with patch.object(
        Dataset,
        "get_active_dataset_status",
        return_value={
            "total": 9999,
            "datasets": {
                "1": {
                    "finished": 9999,
                    "running": 0,
                    "pending": 0,
                    "stages": [
                        {
                            "job_id": "",
                            "stage": "index",
                            "pending": 0,
                            "running": 0,
                            "finished": 9999,
                        }
                    ],
                    "start_time": "2024-06-25T10:58:49.779811",
                    "last_update": "2024-06-25T10:58:49.779819",
                }
            },
        },
    ):
        assert get_task_count(collection_id, redis) == 9999
        assert get_priority(collection_id, redis) in (4, 5, 6)

    with patch.object(
        Dataset,
        "get_active_dataset_status",
        return_value={
            "total": 10001,
            "datasets": {
                "1": {
                    "finished": 10000,
                    "running": 0,
                    "pending": 1,
                    "stages": [
                        {
                            "job_id": "",
                            "stage": "index",
                            "pending": 10001,
                            "running": 0,
                            "finished": 0,
                        }
                    ],
                    "start_time": "2024-06-25T10:58:49.779811",
                    "last_update": "2024-06-25T10:58:49.779819",
                }
            },
        },
    ):
        assert get_task_count(collection_id, redis) == 10001
        assert get_priority(collection_id, redis) in (1, 2, 3)


def test_get_priority_lists():
    redis = get_fakeredis()
    collection_id = 1

    assert Dataset.is_high_prio(redis, collection_id) is False
    assert Dataset.is_low_prio(redis, collection_id) is False

    redis.sadd("tq:prio:low", "1")
    redis.sadd("tq:prio:high", "2")

    assert Dataset.is_low_prio(redis, 1) is True
    assert Dataset.is_high_prio(redis, 1) is False

    assert Dataset.is_low_prio(redis, 2) is False
    assert Dataset.is_high_prio(redis, 2) is True

    assert get_priority(1, redis) == 0
    assert get_priority(2, redis) == 9



import sys
import time
import uuid
import logging
import structlog
from structlog.contextvars import merge_contextvars
from structlog.contextvars import clear_contextvars, bind_contextvars

from servicelayer import settings
from servicelayer.util import unpack_int

LOG_FORMAT_TEXT = "TEXT"
LOG_FORMAT_JSON = "JSON"


def configure_logging(level=logging.INFO):
    """Configure log levels and structured logging"""
    common_processors = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S.%f"),
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
    ]
    if settings.LOG_FORMAT == LOG_FORMAT_TEXT:
        processors = common_processors
        formatter = structlog.stdlib.ProcessorFormatter(
            foreign_pre_chain=processors,
            processor=structlog.dev.ConsoleRenderer(),
        )
    else:
        processors = common_processors + [
            merge_contextvars,
            format_stackdriver,
        ]
        formatter = structlog.stdlib.ProcessorFormatter(
            foreign_pre_chain=processors,
            processor=structlog.processors.JSONRenderer(),
        )

    # configuration for structlog based loggers
    structlog.configure(
        processors=processors
        + [
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
    )

    # handler for low level logs that should be sent to STDOUT
    out_handler = logging.StreamHandler(sys.stdout)
    out_handler.setLevel(level)
    out_handler.addFilter(_MaxLevelFilter(logging.WARNING))
    out_handler.setFormatter(formatter)
    # handler for high level logs that should be sent to STDERR
    error_handler = logging.StreamHandler(sys.stderr)
    error_handler.setLevel(logging.ERROR)
    error_handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    # check to prevent adding duplicate handlers
    if not root_logger.handlers:
        root_logger.addHandler(out_handler)
        root_logger.addHandler(error_handler)


def format_stackdriver(_, __, ed):
    """Stackdriver uses `message` and `severity` keys to display logs"""
    ed["message"] = ed.pop("event")
    ed["severity"] = ed.pop("level", "info").upper()
    return ed


def apply_task_context(task, **kwargs):
    """This clears the current structured logging context and readies it
    for a new task from `servicelayer.jobs`."""
    # Setup context for structured logging
    clear_contextvars()
    bind_contextvars(
        job_id=task.job.id,
        stage=task.stage.stage,
        dataset=task.job.dataset.name,
        start_time=time.time(),
        trace_id=str(uuid.uuid4()),
        retry=unpack_int(task.context.get("retries")),
        **kwargs,
    )


class _MaxLevelFilter(object):
    def __init__(self, highest_log_level):
        self._highest_log_level = highest_log_level

    def filter(self, log_record):
        return log_record.levelno <= self._highest_log_level



import signal
import logging
from timeit import default_timer
import sys
from threading import Thread
from banal import ensure_list
from abc import ABC, abstractmethod

from prometheus_client import start_http_server

from servicelayer import settings
from servicelayer.jobs import Stage
from servicelayer.cache import get_redis
from servicelayer.util import unpack_int
from servicelayer import metrics


log = logging.getLogger(__name__)

# When a worker thread is not blocking, it has to exit if no task is available.
# `TASK_FETCH_RETRY`` determines how many times the worker thread will try to fetch
# a task before quitting.
# `INTERVAL`` determines the interval in seconds between each retry.
INTERVAL = 2
TASK_FETCH_RETRY = 60 / INTERVAL


class Worker(ABC):
    """Workers of all microservices, unite!"""

    def __init__(
        self,
        conn=None,
        stages=None,
        num_threads=settings.WORKER_THREADS,
    ):
        self.conn = conn or get_redis()
        self.stages = stages
        self.num_threads = num_threads
        self.exit_code = 0
        if settings.SENTRY_DSN:
            import sentry_sdk
            from sentry_sdk.integrations.threading import ThreadingIntegration

            sentry_sdk.init(
                dsn=settings.SENTRY_DSN,
                traces_sample_rate=0,
                release=settings.SENTRY_RELEASE,
                environment=settings.SENTRY_ENVIRONMENT,
                send_default_pii=False,
                disabled_integrations=[
                    ThreadingIntegration(),
                ],
            )

    def _handle_signal(self, signal, frame):
        log.warning(f"Shutting down worker (signal {signal})")
        self.exit_code = int(signal)
        # Exit eagerly without waiting for current task to finish running
        sys.exit(self.exit_code)

    def handle_safe(self, task):
        retries = unpack_int(task.context.get("retries"))

        try:
            metrics.TASKS_STARTED.labels(stage=task.stage.stage).inc()
            start_time = default_timer()
            self.handle(task)
            duration = max(0, default_timer() - start_time)
            metrics.TASK_DURATION.labels(stage=task.stage.stage).observe(duration)
            metrics.TASKS_SUCCEEDED.labels(
                stage=task.stage.stage, retries=retries
            ).inc()
        except SystemExit as exc:
            self.exit_code = exc.code
            self.retry(task)
        except KeyboardInterrupt:
            self.exit_code = 23
            self.retry(task)
        except Exception:
            if 0 == self.exit_code:
                self.exit_code = 23
            self.retry(task)
            log.exception("Error in task handling")
        finally:
            task.done()
            self.after_task(task)

    def init_internal(self):
        self.exit_code = 0
        self.boot()

    def run_prometheus_server(self):
        if not settings.PROMETHEUS_ENABLED:
            return

        def run_server():
            port = settings.PROMETHEUS_PORT
            log.info(f"Running Prometheus metrics server on port {port}")
            start_http_server(port)

        thread = Thread(target=run_server)
        thread.start()
        thread.join()

    def retry(self, task):
        retries = unpack_int(task.context.get("retries"))
        if retries < settings.WORKER_RETRY:
            retry_count = retries + 1
            log.warning(
                f"Queueing failed task for retry #{retry_count}/{settings.WORKER_RETRY}..."  # noqa
            )
            metrics.TASKS_FAILED.labels(
                stage=task.stage.stage,
                retries=retries,
                failed_permanently=False,
            ).inc()
            task.context["retries"] = retry_count
            task.stage.queue(task.payload, task.context)
        else:
            log.warning(
                f"Failed task, exhausted retry count of {settings.WORKER_RETRY}"
            )
            metrics.TASKS_FAILED.labels(
                stage=task.stage.stage,
                retries=retries,
                failed_permanently=True,
            ).inc()

    def process(self, blocking=True, interval=INTERVAL):
        retries = 0
        while retries <= TASK_FETCH_RETRY:
            if self.exit_code > 0:
                log.info("Worker thread is exiting")
                return self.exit_code
            self.periodic()
            stages = self.get_stages()
            task = Stage.get_task(self.conn, stages, timeout=interval)
            if task is None:
                if not blocking:
                    # If we get a null task, retry to fetch a task
                    # a bunch of times before quitting
                    if retries >= TASK_FETCH_RETRY:
                        log.info("Worker thread is exiting")
                        return self.exit_code
                    else:
                        retries += 1
                continue
            # when we get a good task, reset retry count
            retries = 0
            self.handle_safe(task)

    def sync(self):
        """Process only the tasks already in the job queue, but do not
        go into an infinte loop waiting for new ones."""
        self.init_internal()
        self.process(blocking=False, interval=None)

    def run(self, blocking=True, interval=INTERVAL):
        signal.signal(signal.SIGINT, self._handle_signal)
        signal.signal(signal.SIGTERM, self._handle_signal)
        self.init_internal()
        self.run_prometheus_server()

        def process():
            return self.process(blocking=blocking, interval=interval)

        if not self.num_threads:
            return process()
        log.info("Worker has %d threads.", self.num_threads)
        threads = []
        for _ in range(self.num_threads):
            thread = Thread(target=process)
            thread.daemon = True
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()
        return self.exit_code

    def get_stages(self):
        """Easily allow the user to make the active stages dynamic."""
        return self.stages

    def boot(self):
        """Optional hook for the boot-up of the worker."""
        pass

    def periodic(self):
        """Optional hook for running a periodic task checker."""
        pass

    def after_task(self, task):
        """Optional hook excuted after handling a task"""
        pass

    def dispatch_pipeline(self, task, payload):
        """Some queues use a continuation passing style pipeline argument
        to specify the next steps for a processing chain."""
        pipeline = ensure_list(task.context.get("pipeline"))
        if len(pipeline) == 0:
            return
        next_stage = pipeline.pop(0)
        stage = task.job.get_stage(next_stage)
        context = dict(task.context)
        context["pipeline"] = pipeline
        stage.queue(payload, context)

    @abstractmethod
    def handle(self, task):
        raise NotImplementedError



from os import environ
from banal.bools import as_bool


def get(name, default=None):
    value = environ.get(name)
    if value is not None:
        return str(value)
    if default is not None:
        return str(default)


def to_int(name, default=0):
    """Extract an integer from the environment."""
    try:
        return int(get(name))
    except (TypeError, ValueError):
        return default


def to_bool(name, default=False):
    """Extract a boolean value from the environment consistently."""
    return as_bool(get(name), default=default)


def to_list(name, default=[], separator=":"):
    """Extract a list of values from the environment consistently.
    Multiple values are by default expected to be separated by a colon (':'),
    like in the UNIX $PATH variable.
    """
    value = get(name)
    if value is None:
        return list(default)
    return [e.strip() for e in value.split(separator)]
